// dependance of context.h
#include <gcc-plugin.h>
// 'g' global variable
#include <context.h>
// dependance of gimple.h
#include <tree.h>
// dependance of gimple-iterator.h
#include <gimple.h>
// gsi
#include <gimple-iterator.h>
//bitmaps
#include <bitmap.h>
// basename
#include <filesystem>
// list
#include <list>

// our pass
#include "pass_mpi_collective.hpp"

// Definitions
const pass_data my_pass_data =
{
	GIMPLE_PASS, /* type */
	"MPI COLLECTIVE", /* name */
	OPTGROUP_NONE, /* optinfo_flags */
	TV_OPTIMIZE, /* tv_id */
	0, /* properties_required */
	0, /* properties_provided */
	0, /* properties_destroyed */
	0, /* todo_flags_start */
	0, /* todo_flags_finish */
};

// Class functions
pass_mpi_collective::pass_mpi_collective(gcc::context *ctxt)
	: opt_pass(my_pass_data, ctxt)
{
	for (int i = 0; i < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE; ++i)
	{
		collective_max_rank[i] = 0;
	}
}

pass_mpi_collective *pass_mpi_collective::clone()
{
	return new pass_mpi_collective(g);
}

bool pass_mpi_collective::gate(function *fun)
{
	//printf("In gate of: %s\n", fndecl_name(fun->decl));
	printf("In gate of: %s\n", function_name(fun));
	//printf("In gate of: %s\n", current_function_name());
	return true;
}

unsigned int pass_mpi_collective::execute(function *fun)
{
	printf("In execute of: %s\n", function_name(fun));

	print_tree(fun);
	split_blocks(fun);
	printf("\t--------------------------[split]---------------------------\n");
	print_tree(fun);

	alloc_bb_aux(fun);
	label_collec(fun);
	label_dom(fun);
	label_dom_front(fun);

	alloc_edge_aux(fun);

	mark_edge(fun);
	rank_collective(fun);
	//better_rank_collective(fun);

	//bitmap_head pdbitmap = get_set_mpi_coll_rank(fun, 2, MPI_BARRIER);

	//debug_bitmap(&pdbitmap);

	//bitmap_head pdf = get_frontier_from_pds(fun, &pdbitmap);
	//debug_bitmap(&pdf);

	cfgviz_dump(fun, "_split");

	printf("[][][][][][][][][][]\n");
	raise_warning_if_mpi_wrong(fun);

	free_dom_data();
	free_bb_aux(fun);
	free_edge_aux(fun);

	return 0;
}

void pass_mpi_collective::print_tree(function *fun)
{
	basic_block bb;
	gimple_stmt_iterator gsi;
	gimple *stmt;
	//size_t line;

	FOR_EACH_BB_FN(bb, fun)
	{
		printf("\tBasic block: %d\n", bb->index);
		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi))
		{
			stmt = gsi_stmt(gsi);
			//line = gimple_lineno(stmt);
			//printf("\tstatement at line = %lu\n", line);
			if (is_func(stmt))
			{
				printf("\t\tStatement is a function call (%s)\n",
				       fndecl_name(gimple_call_fndecl(stmt)));
				if (__is_mpi_func(stmt))
				{
					printf("\t\t\tStatement is a MPI function call (%s)\n",
					       fndecl_name(gimple_call_fndecl(stmt)));
					if (__is_mpi_collec(stmt)
					    != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
					{
						printf("\t\t\t\tStatement is a MPI collective (%s)\n",
						       fndecl_name(gimple_call_fndecl(stmt)));
					}
				}
			}
		}
	}
}

// MPI function / collectives detections

bool pass_mpi_collective::is_func(gimple *stmt)
{
	return is_gimple_call(stmt);
}

bool pass_mpi_collective::__is_mpi_func(gimple *stmt)
{
	return strncmp(fndecl_name(gimple_call_fndecl(stmt)), "MPI_", 4) == 0;
}

bool pass_mpi_collective::is_mpi_func(gimple *stmt)
{
	return is_func(stmt) && __is_mpi_func(stmt);
}

mpi_collective_code pass_mpi_collective::__is_mpi_collec(gimple *stmt)
{
	size_t i;
	const char *callee_name;

	callee_name = fndecl_name(gimple_call_fndecl(stmt));
	for (i = 0; i < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE; ++i)
	{
		if (strcmp(mpi_collective_name[i], callee_name) == 0)
		{
			return (enum mpi_collective_code) i;
		}
	}
	return LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
}

mpi_collective_code pass_mpi_collective::is_mpi_collec(gimple *stmt)
{
	if (!is_mpi_func(stmt))
	{
		return LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
	}
	return __is_mpi_collec(stmt);
}

// split blocks according to mpi collectives

void pass_mpi_collective::split_blocks(function *fun)
{
	basic_block bb;
	gimple_stmt_iterator gsi;
	gimple *stmt;
	gimple *prev_stmt;
	size_t nb_collective;
	bool split;

	FOR_EACH_BB_FN(bb, fun)
	{
		nb_collective = 0;
		stmt = NULL;
		split = false;
		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi) && !split; gsi_next(&gsi))
		{
			prev_stmt = stmt;
			stmt = gsi_stmt(gsi);
			if (is_mpi_collec(stmt) != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
			{
				++nb_collective;
			}
			if (nb_collective >= 2)
			{
				split_block(bb, prev_stmt);
				split = true;
				//The inner for doesn't stop naturally, whereas the
				//FOR_EACH_BB_FN *will* iterate on the new bb we just split.
				//hence the split=true
			}
		}
	}
}

void pass_mpi_collective::alloc_bb_aux(function *fun)
{
	basic_block bb;
	
	FOR_ALL_BB_FN(bb, fun)
	{
		bb->aux = (bb_data *) new bb_data();
		bb_data *data = ((bb_data *) bb->aux);
		data->mpi_code = LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
		bitmap_initialize(&data->dom, &bitmap_default_obstack);	
		bitmap_initialize(&data->post_dom, &bitmap_default_obstack);
		bitmap_initialize(&data->dom_front, &bitmap_default_obstack);
		bitmap_initialize(&data->post_dom_front, &bitmap_default_obstack);
		for (int i = 0; i < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE; ++i)
		{
			data->collective_rank[i] = 0;
		}
		bitmap_initialize(&data->seens, &bitmap_default_obstack);
	}

	reset_bb_mark(fun);
}

// gestions des aux
void pass_mpi_collective::label_collec(function *fun)
{
	basic_block bb;
	gimple_stmt_iterator gsi;
	gimple *stmt;
	mpi_collective_code mpi_code;

	FOR_ALL_BB_FN(bb, fun)
	{
		((bb_data *) bb->aux)->mpi_code = LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi))
		{
			stmt = gsi_stmt(gsi);
			mpi_code = is_mpi_collec(stmt);
			if (mpi_code != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
			{
				((bb_data *) bb->aux)->mpi_code = mpi_code;
			}
		}
	}
}

void pass_mpi_collective::free_bb_aux(function *fun)
{
	basic_block bb;
	FOR_ALL_BB_FN(bb, fun)
	{
		if (bb->aux != NULL)
		{
			bb_data *data = ((bb_data *) bb->aux);
			bitmap_clear(&data->dom);
			bitmap_clear(&data->post_dom);
			bitmap_clear(&data->dom_front);
			bitmap_clear(&data->post_dom_front);
			bitmap_clear(&data->seens);
			delete (bb_data *) bb->aux;
			bb->aux = NULL;
		}
	}
}

// calculate dominance
void pass_mpi_collective::label_dom(function *fun)
{
	basic_block bb;
	auto_vec<basic_block> dom_gccvec;
	auto_vec<basic_block> post_dom_gccvec;
	size_t size_dom;
	size_t size_post_dom;

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	FOR_ALL_BB_FN(bb, fun)
	{
		bb_data *data = ((bb_data *) bb->aux);

		dom_gccvec = get_all_dominated_blocks(CDI_DOMINATORS, bb);
		size_dom = dom_gccvec.length();
		for (size_t i = 0; i < size_dom; ++i)
		{
			if (dom_gccvec[i]->index != bb->index)
			{
				bitmap_set_bit(&data->dom, dom_gccvec[i]->index);
			}
		}

		post_dom_gccvec = get_all_dominated_blocks(CDI_POST_DOMINATORS, bb);
		size_post_dom = post_dom_gccvec.length();
		for (size_t i = 0; i < size_post_dom; ++i)
		{
			if (post_dom_gccvec[i]->index != bb->index)
			{
				bitmap_set_bit(&data->post_dom, post_dom_gccvec[i]->index);
			}
		}
	}
}

void pass_mpi_collective::label_dom_front(function *fun)
{
	edge p;
	edge_iterator ei;
	basic_block bb;
	basic_block dom_bb;
	basic_block runner;
	bb_data *data;

	FOR_EACH_BB_FN(bb, fun)
	{
		if (EDGE_COUNT(bb->preds) < 2)
		{
			continue;
		}
		dom_bb = get_immediate_dominator(CDI_DOMINATORS, bb);
		FOR_EACH_EDGE(p, ei, bb->preds)
		{
			runner = p->src;
			if (runner == ENTRY_BLOCK_PTR_FOR_FN(fun))
			{
				continue;
			}
			while (runner != dom_bb)
			{
				data = (bb_data *) runner->aux;
				if (bitmap_bit_p(&data->dom_front, bb->index))
				{
					break;
				}
				bitmap_set_bit(&data->dom_front, bb->index);
				runner = get_immediate_dominator(CDI_DOMINATORS, runner);
			}
		}
	}

	FOR_EACH_BB_FN(bb, fun)
	{
		if (EDGE_COUNT(bb->succs) < 2)
		{
			continue;
		}
		dom_bb = get_immediate_dominator(CDI_POST_DOMINATORS, bb);
		FOR_EACH_EDGE(p, ei, bb->succs)
		{
			runner = p->dest;
			if (runner == EXIT_BLOCK_PTR_FOR_FN(fun))
			{
				continue;
			}
			while (runner != dom_bb)
			{
				data = (bb_data *) runner->aux;
				if (bitmap_bit_p(&data->post_dom_front, bb->index))
				{
					break;
				}
				bitmap_set_bit(&data->post_dom_front, bb->index);
				runner = get_immediate_dominator(CDI_POST_DOMINATORS, runner);
			}
		}
	}
}

void pass_mpi_collective::free_dom_data()
{
	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
}

// parkour de graph
void pass_mpi_collective::reset_bb_mark(function *fun)
{
	basic_block bb;

	FOR_ALL_BB_FN(bb, fun)
	{
		((bb_data *) bb->aux)->mark1 = 0;
		((bb_data *) bb->aux)->mark2 = 0;
	}
}

void pass_mpi_collective::alloc_edge_aux(function *fun)
{
	basic_block bb;
	edge e;
	edge_iterator ei;

	FOR_ALL_BB_FN(bb, fun)
	{
		FOR_EACH_EDGE(e, ei, bb->succs)
		{
			e->aux = new edge_data();
		}
	}
}

void pass_mpi_collective::free_edge_aux(function *fun)
{
	basic_block bb;
	edge e;
	edge_iterator ei;

	FOR_ALL_BB_FN(bb, fun)
	{
		FOR_EACH_EDGE(e, ei, bb->succs)
		{
			if (e->aux != NULL)
			{
				delete (edge_data *) e->aux;
				e->aux = NULL;
			}
		}
	}
}

// naive version of mark_edge (work but Omax(2^n))
void pass_mpi_collective::__mark_edge_naif(function *fun)
{
	__mark_edge_naif(ENTRY_BLOCK_PTR_FOR_FN(fun));
}

void pass_mpi_collective::__mark_edge_naif(basic_block bb)
{
	edge e;
	edge_iterator ei;

	((bb_data *) bb->aux)->mark1 = 1;
	FOR_EACH_EDGE(e, ei, bb->succs)
	{
		if (((bb_data *) e->dest->aux)->mark1)
		{
			((edge_data *) e->aux)->loop = true;
		}
		else
		{
			__mark_edge_naif(e->dest);
		}
	}
	((bb_data *) bb->aux)->mark1 = 0;
}

void pass_mpi_collective::__mark_edge(function *fun)
{

	edge e1;
	edge_iterator ei1;
	std::list<basic_block> nodes1;
	basic_block bb;

	edge e2;
	edge_iterator ei2;
	std::list<basic_block> nodes2;
	basic_block cursor;

	int index;

	index = 0;

	nodes1.push_back(ENTRY_BLOCK_PTR_FOR_FN(fun));
	((bb_data *) ENTRY_BLOCK_PTR_FOR_FN(fun)->aux)->mark1 = 1;

	while (!nodes1.empty())
	{
		bb = nodes1.front();
		nodes1.pop_front();

		FOR_EACH_EDGE(e1, ei1, bb->succs)
		{
			if (((bb_data *) e1->dest->aux)->mark1)
			{
				continue;
			}
			((bb_data *) bb->aux)->mark1 = 1;
			nodes1.push_back(e1->dest);
		}

		if (EDGE_COUNT(bb->preds) < 2)
		{
			continue;
		}

		++index;
		nodes2.clear();
		nodes2.push_back(bb);
		((bb_data *) bb->aux)->mark2 = index;

		while (!nodes2.empty())
		{
			cursor = nodes2.front();
			nodes2.pop_front();

			FOR_EACH_EDGE(e2, ei2, cursor->succs)
			{
				if (e2->dest == bb)
				{
					((edge_data *) e2->aux)->loop = true;
				}
				if (((edge_data *) e2->aux)->loop)
				{
					continue;
				}

				if (((bb_data *) e2->dest->aux)->mark2 >= index)
				{
					continue;
				}
				((bb_data *) cursor->aux)->mark2 = index;
				nodes2.push_back(e2->dest);
			}
		}
	}
	reset_bb_mark(fun);
}

void pass_mpi_collective::mark_edge(function *fun)
{
	mark_edge(ENTRY_BLOCK_PTR_FOR_FN(fun));
	reset_bb_mark(fun);
}

void pass_mpi_collective::mark_edge(basic_block bb)
{
	edge e;
	edge_iterator ei;

	((bb_data *) bb->aux)->mark1 = 1;
	bitmap_set_bit(&((bb_data *) bb->aux)->seens, bb->index);
	FOR_EACH_EDGE(e, ei, bb->succs)
	{
		if (((edge_data *) e->aux)->loop)
		{
			continue;
		}
		if (bitmap_bit_p(&((bb_data *) bb->aux)->seens, e->dest->index))
		{
			((edge_data *) e->aux)->loop = true;
		}
		if (! ((bb_data *) e->dest->aux)->mark1)
		{
			bitmap_ior_into(&((bb_data *) e->dest->aux)->seens,
			                &((bb_data *) bb->aux)->seens);
			mark_edge(e->dest);
		}
	}
}

void pass_mpi_collective::rank_collective(function *fun)
{
	__rank_collective(EXIT_BLOCK_PTR_FOR_FN(fun));
	reset_bb_mark(fun);
}

void pass_mpi_collective::__rank_collective(basic_block bb)
{
	edge e;
	edge_iterator ei;

	((bb_data *) bb->aux)->mark1 = 1;
	FOR_EACH_EDGE(e, ei, bb->preds)
	{
		if (((edge_data *) e->aux)->loop)
		{
			continue;
		}

		if (! ((bb_data *) e->src->aux)->mark1)
		{
			__rank_collective(e->src);
		}

		for (int mpi_code = 0;
		     mpi_code < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
		     ++mpi_code)
		{
			((bb_data *) bb->aux)->collective_rank[mpi_code] =
			  std::max(((bb_data *) bb->aux)->collective_rank[mpi_code],
			           ((bb_data *) e->src->aux)->collective_rank[mpi_code]);
			if (((bb_data *) bb->aux)->mpi_code == mpi_code)
			{
				((bb_data *) bb->aux)->collective_rank[mpi_code]++;
				collective_max_rank[mpi_code] = std::max(
				                                  collective_max_rank[mpi_code],
				                                  ((bb_data *) bb->aux)->collective_rank[mpi_code]);
			}
		}
	}
}

void pass_mpi_collective::better_rank_collective(function *fun)
{
	size_t i;
	int next_rank = 0;

	for (i = 0; i < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE; ++i)
	{
		next_rank = __better_rank_collective(EXIT_BLOCK_PTR_FOR_FN(fun),
		                                     (enum mpi_collective_code) i,
		                                     next_rank);
	}
	reset_bb_mark(fun);
}

int pass_mpi_collective::__better_rank_collective(basic_block bb,
  mpi_collective_code mpi_code, int rank)
{
	edge e;
	edge_iterator ei;
	int next_rank;

	next_rank = rank;
	((bb_data *) bb->aux)->mark1 = mpi_code + 1;

	FOR_EACH_EDGE(e, ei, bb->preds)
	{
		if (((edge_data *) e->aux)->loop)
		{
			continue;
		}

		// cache or travell
		if (((bb_data *) e->src->aux)->mark1 <= mpi_code)
		{
			next_rank =
			  std::max(next_rank,
			           __better_rank_collective(e->src, mpi_code, rank));
		}
		else
		{
			next_rank =
			  std::max(next_rank,
			           ((bb_data *) e->src->aux)->collective_rank[mpi_code]);
		}
	}
	// set mpi code
	if (((bb_data *) bb->aux)->mpi_code == mpi_code)
	{
		next_rank++;
	}
	((bb_data *) bb->aux)->collective_rank[mpi_code] = next_rank;
	return next_rank;
}

bool pass_mpi_collective::__is_pd_set_mpi_coll_rank(function *fun,
  basic_block bb, int rank, int mpi_code, bitmap pds)
{
	edge e;
	edge_iterator ei;
	bool is_pd_by_set;//is postdominated by set
	int count_edges;// nb. of succs that are not loop.

	//printf("entered bb->index : %i\n",bb->index);
	((bb_data *) bb->aux)->mark1 = 1;

	// if we are on an MPI collective of the requiered rank.
	if (((bb_data *) bb->aux)->mpi_code == mpi_code
	    && ((bb_data *) bb->aux)->collective_rank[mpi_code]
	    == rank) //if we are in the pd set
	{
		//printf("They are my people, %i is 1\n", bb->index);
		bitmap_set_bit(pds, bb->index);
		return true;
	}

	is_pd_by_set = true;
	count_edges  = 0;

	FOR_EACH_EDGE(e, ei, bb->succs)
	{
		if (((edge_data *) e->aux)->loop)
		{
			//printf("loop detected... Jumping out...\n");
			continue; //if marked as a loop or already visited, ignore
		}

		count_edges++;
		//printf("%i: %i\n",bb->index, count_edges);
		// cache or travell
		if (((bb_data *) e->dest->aux)->mark1)
		{
			//printf("node %i already visited.. jumping out...\n",
			//	e->dest->index);
			is_pd_by_set = is_pd_by_set && bitmap_bit_p(pds, e->dest->index);
		}
		else
		{
			is_pd_by_set =
			  __is_pd_set_mpi_coll_rank(fun, e->dest, rank, mpi_code, pds)
			  && is_pd_by_set;
		}
	}

	if (count_edges == 0)
	{
		//printf("Leaf ; %i is 0\n", bb->index);
		return false; //we are in a leaf
	}

	if (is_pd_by_set)
	{
		// add the bb to the pdset if it is post dominated by set
		bitmap_set_bit(pds, bb->index);
	}

	//printf("I now know that %i is %b\n", bb->index, is_pd_by_set);
	return is_pd_by_set;
}

void pass_mpi_collective::get_set_mpi_coll_rank(function *fun,
  int rank, int mpi_code, bitmap pds)
{
	__is_pd_set_mpi_coll_rank(fun, ENTRY_BLOCK_PTR_FOR_FN(fun),
	                          rank, mpi_code, pds);
	reset_bb_mark(fun);
}

void pass_mpi_collective::get_frontier_from_pds(function *fun,
		bitmap pds, bitmap pdf)
{
	//we check each bb, if a bb is in the postdom set, we check its
	//most adjacents preds. If a pred is not in the postdom set, it
	//is a elm of the frontier of set.
	basic_block bb;
	edge e;
	edge_iterator ei;

	FOR_ALL_BB_FN(bb, fun)
	{
		if (bitmap_bit_p(pds, bb->index))
		{
			//printf("entering %i\n", bb->index);
			FOR_EACH_EDGE(e, ei, bb->preds)
			{
				//printf("\tchecking out my parent %i\n", e->src->index);
				if (! ((edge_data *) e->aux)->loop && !bitmap_bit_p(pds, e->src->index))
				{
					bitmap_set_bit(pdf, e->src->index);
				}
			}
		}
	}
}

void pass_mpi_collective::raise_warning_if_mpi_wrong(function *fun)
{
	for (int mpi_code = 0; mpi_code < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
	     ++mpi_code)
	{
		bitmap_head pds_head;
		bitmap_head pdf_head;
		bitmap pds = &pds_head;
		bitmap pdf = &pdf_head;
		bitmap tmp = NULL;
	
		bitmap_initialize(pds, &bitmap_default_obstack);
		bitmap_initialize(pdf, &bitmap_default_obstack);
		int stop = 0;

		for (int i = 1 ; i < collective_max_rank[mpi_code] + 1 ; i++)
		{
			printf("mpi_code: %i at rank: %i\n", mpi_code, i);
			// TODO: split in 2, get mpi set & find set post dom 
			get_set_mpi_coll_rank(fun, i, mpi_code, pds);

			while(! bitmap_empty_p(pds)){
				stop++;
				if(stop > 10)
					return;
				get_frontier_from_pds(fun, pds, pdf);
				
				printf("----\n");
				debug_bitmap(pds);
				debug_bitmap(pdf);

				bitmap_clear(pds);
				
				// swap pds & pdf
				{
					tmp = pdf;
					pdf = pds;
					pds = tmp;
				}
				// pds = 
			}
		}
	}
}

//============================	GRAPHVIZ  =====================================

/* Build a filename (as a string) based on function name */
char *pass_mpi_collective::cfgviz_generate_filename(function *fun,
  const char *suffix)
{
	char *target_filename;
	int line;

	target_filename = (char *)xmalloc(1024 * sizeof(char));
	line = LOCATION_LINE(fun->function_start_locus);
	const char *mfilename =
	  std::filesystem::path(LOCATION_FILE(fun->function_start_locus))
	  .filename().c_str();

	snprintf(target_filename, 1024, "%s%s_%s_%d_%s.dot",
	         "dot/",
	         mfilename,
	         function_name(fun),
	         line,
	         suffix);

	return target_filename;
}

/* Dump the graphviz representation of function 'fun' in file 'out' */
void pass_mpi_collective::cfgviz_internal_dump(function *fun, FILE *out)
{
	basic_block bb;
	edge_iterator eit;
	edge e;

	bitmap_iterator bi;
	unsigned int index;

	fprintf(out, "Digraph G{\n");
	FOR_ALL_BB_FN(bb, fun)
	{
		bb_data *data = ((bb_data *) bb->aux);

		fprintf(out, "%d [label=\"BB %d", bb->index, bb->index);
		if (data->mpi_code != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
		{
			fprintf(out, "|%s (%d) ", mpi_collective_name[data->mpi_code],
			        data->collective_rank[data->mpi_code]);
		}

		if (! bitmap_empty_p(&data->dom))
		{
			fprintf(out, "|dom(");
			
			EXECUTE_IF_SET_IN_BITMAP(&data->dom, 0, index, bi){
					fprintf(out, "%d,", index);
			}
			fprintf(out, ")");
		}

		if (! bitmap_empty_p(&data->post_dom))
		{
			fprintf(out, "|post_dom(");
			EXECUTE_IF_SET_IN_BITMAP(&data->post_dom, 0, index, bi){
				fprintf(out, "%d,", index);
			}
			fprintf(out, ")");
		}

		if (! bitmap_empty_p(&data->dom_front))
		{
			fprintf(out, "|dom_front(");
			EXECUTE_IF_SET_IN_BITMAP(&data->dom_front, 0, index, bi){
				fprintf(out, "%d,", index);
			}
			fprintf(out, ")");
		}

		if (! bitmap_empty_p(&data->post_dom_front))
		{
			fprintf(out, "|post_dom_front(");
			EXECUTE_IF_SET_IN_BITMAP(&data->post_dom_front, 0, index, bi){
				fprintf(out, "%d,", index);
			}
			fprintf(out, ")");
		}

		fprintf(out, "\" shape=ellipse]\n");

		FOR_EACH_EDGE(e, eit, bb->succs)
		{
			const char *label = "";
			if (e->flags == EDGE_TRUE_VALUE)
			{
				label = "true";
			}
			else if (e->flags == EDGE_FALSE_VALUE)
			{
				label = "false";
			}

			if (((edge_data *) e->aux)->loop)
			{
				label = "remove_loop";
			}

			fprintf(out, "%d -> %d [color=red label=\"%s\"]\n",
			        bb->index, e->dest->index, label);
		}
	}
	fprintf(out, "}\n");
}

void pass_mpi_collective::cfgviz_dump(function *fun, const char *suffix)
{
	char *target_filename;
	FILE *out;

	target_filename = cfgviz_generate_filename(fun, suffix);

	printf("[GRAPHVIZ] Generating CFG of function %s in file <%s>\n",
	       current_function_name(), target_filename);

	out = fopen(target_filename, "w");

	cfgviz_internal_dump(fun, out);

	fclose(out);
	free(target_filename);
}
